"use strict";
var __rest = (this && this.__rest) || function (s, e) {
    var t = {};
    for (var p in s) if (Object.prototype.hasOwnProperty.call(s, p) && e.indexOf(p) < 0)
        t[p] = s[p];
    if (s != null && typeof Object.getOwnPropertySymbols === "function")
        for (var i = 0, p = Object.getOwnPropertySymbols(s); i < p.length; i++) {
            if (e.indexOf(p[i]) < 0 && Object.prototype.propertyIsEnumerable.call(s, p[i]))
                t[p[i]] = s[p[i]];
        }
    return t;
};
Object.defineProperty(exports, "__esModule", { value: true });
const parser_1 = require("./parser");
const types_1 = require("./types");
const helpers_1 = require("./helpers");
// --------------------------------------------------------------------- //
// ------------------------------- HELPER ------------------------------ //
// --------------------------------------------------------------------- //
function isParameterProperties(meta) {
    return meta && meta.kind === "Property" && meta.isParameter;
}
function getDecorators(targetClass, targetType, target, index) {
    const natives = Reflect.getOwnMetadata(types_1.DECORATOR_KEY, targetClass) || [];
    const result = [];
    for (let _a of natives) {
        const { allowMultiple, inherit } = _a, item = __rest(_a, ["allowMultiple", "inherit"]);
        const par = item;
        if (item.targetType === targetType && item.target === target && (index == undefined || par.targetIndex === index)) {
            result.push(Object.assign(Object.assign({}, item.value), { [types_1.DecoratorOptionId]: { allowMultiple, inherit } }));
        }
    }
    return result;
}
function getParents(target) {
    return target.prototype ? [...getParents(Object.getPrototypeOf(target)), target] : [Object];
}
function createVisitors(...visitors) {
    return (value, ctx) => {
        return visitors.reduce((a, b) => !!a ? b(a, ctx) : a, value);
    };
}
exports.createVisitors = createVisitors;
function purifyTraversal(meta, ctx) {
    const result = ctx.visitor(meta, ctx);
    for (const key in result) {
        if (["parameters", "properties", "methods", "ctor"].some(x => x === key)) {
            const item = result[key];
            if (Array.isArray(item)) {
                const node = item.map((x, i) => purifyTraversal(x, Object.assign(Object.assign({}, ctx), { parent: result })));
                result[key] = node.filter(x => !!x);
            }
            else {
                const node = purifyTraversal(item, Object.assign(Object.assign({}, ctx), { parent: item }));
                result[key] = node;
            }
        }
    }
    return result;
}
exports.purifyTraversal = purifyTraversal;
// --------------------------------------------------------------------- //
// ------------------------------ EXTENDER ----------------------------- //
// --------------------------------------------------------------------- //
function traverseArray(childItem, parentItem) {
    const items = [];
    // iterate all child nodes
    for (let i = 0; i < childItem.length; i++) {
        const ch = childItem[i];
        const prn = parentItem.find(x => x.name === ch.name);
        items.push(traverse(ch, prn));
    }
    // iterate only non existed parent nodes
    for (const prn of parentItem) {
        const ch = childItem.find(x => x.name === prn.name);
        if (!ch)
            items.push(traverse(ch, prn));
    }
    return items.filter(x => !!x);
}
function traverse(child, parent) {
    const result = extendsNode(child, parent);
    for (const key in child) {
        if (["parameters", "properties", "methods", "ctor"].some(x => x === key)) {
            const childItem = child[key];
            if (Array.isArray(childItem)) {
                const parentItem = (parent || {})[key] || [];
                result[key] = traverseArray(childItem, parentItem);
            }
            else {
                const parentItem = parent[key];
                const node = traverse(childItem, parentItem);
                result[key] = node;
            }
        }
    }
    return result;
}
function extendsNode(child, parent) {
    // don't extends constructor
    if (child && child.kind === "Constructor")
        return child;
    // parent parameter should not be copied
    if (!child && parent.kind === "Parameter")
        return undefined;
    // parameter property: copy parent index to get proper design type information
    if (isParameterProperties(child) && isParameterProperties(parent))
        return Object.assign(Object.assign({}, child), { owner: [...parent.owner, ...child.owner], index: parent.index });
    // extends owner for all node except constructor
    if (child && parent && parent.kind !== "Constructor")
        return Object.assign(Object.assign({}, child), { owner: [...parent.owner, ...child.owner] });
    // if child not exists, by default just copy parent
    return child !== null && child !== void 0 ? child : parent;
}
function extendsClass(meta) {
    const objectGraph = {
        kind: "Class", type: Object, name: "Object",
        ctor: { kind: "Constructor", name: "constructor", parameters: [] },
        methods: [], properties: [], decorators: [],
        super: Object,
        owner: [Object]
    };
    const parent = Object.getPrototypeOf(meta.type);
    const parentMeta = parent.prototype ? extendsClass(parser_1.parseClass(parent)) : objectGraph;
    return traverse(meta, parentMeta);
}
exports.extendsClass = extendsClass;
// --------------------------------------------------------------------- //
// ------------------------- PURIFIER VISITORS ------------------------- //
// --------------------------------------------------------------------- //
var visitors;
(function (visitors) {
    function addSuperclassMeta(meta, ctx) {
        if (meta.kind === "Class")
            return extendsClass(meta);
        return meta;
    }
    visitors.addSuperclassMeta = addSuperclassMeta;
    function addsDesignTypes(meta, ctx) {
        const getType = (type, i) => type[i] === Array ? [Object] : type[i];
        if (meta.kind === "Method") {
            const returnType = Reflect.getOwnMetadata(types_1.DESIGN_RETURN_TYPE, meta.owner[0].prototype, meta.name);
            return Object.assign(Object.assign({}, meta), { returnType });
        }
        else if (isParameterProperties(meta)) {
            const parTypes = Reflect.getOwnMetadata(types_1.DESIGN_PARAMETER_TYPE, meta.owner[0]) || [];
            return Object.assign(Object.assign({}, meta), { type: getType(parTypes, meta.index) });
        }
        else if (meta.kind === "Property") {
            const type = Reflect.getOwnMetadata(types_1.DESIGN_TYPE, meta.owner[0].prototype, meta.name);
            return Object.assign(Object.assign({}, meta), { type });
        }
        else if (meta.kind === "Parameter" && ctx.parent.kind === "Constructor") {
            const parTypes = Reflect.getOwnMetadata(types_1.DESIGN_PARAMETER_TYPE, meta.owner[0]) || [];
            return Object.assign(Object.assign({}, meta), { type: getType(parTypes, meta.index) });
        }
        else if (meta.kind === "Parameter" && ctx.parent.kind === "Method") {
            const parTypes = Reflect.getOwnMetadata(types_1.DESIGN_PARAMETER_TYPE, meta.owner[0].prototype, ctx.parent.name) || [];
            return Object.assign(Object.assign({}, meta), { type: getType(parTypes, meta.index) });
        }
        else
            return meta;
    }
    visitors.addsDesignTypes = addsDesignTypes;
    function extendDecorators(ownDecorators, parentDecorators) {
        const result = [...ownDecorators];
        for (const decorator of parentDecorators) {
            const options = decorator[types_1.DecoratorOptionId];
            // continue, if the decorator is not inheritable
            if (!options.inherit)
                continue;
            // continue, if allow multiple and already has decorator with the same ID
            if (!options.allowMultiple && ownDecorators.some(x => x[types_1.DecoratorId] === decorator[types_1.DecoratorId]))
                continue;
            result.push(decorator);
        }
        return result;
    }
    function addsDecorators(meta, ctx) {
        const flat = (d) => [].concat(...d);
        const getOwners = (owners) => [ctx.target, owners.filter(x => x !== ctx.target)];
        if (meta.kind === "Parameter" || isParameterProperties(meta)) {
            const targetName = meta.kind === "Parameter" ? ctx.parent.name : "constructor";
            const [owner, parents] = getOwners(meta.owner);
            const ownDecorators = getDecorators(owner, "Parameter", targetName, meta.index);
            const parentDecorators = parents.map(x => getDecorators(x, "Parameter", targetName, meta.index));
            return Object.assign(Object.assign({}, meta), { decorators: extendDecorators(ownDecorators, flat(parentDecorators)) });
        }
        else if (meta.kind === "Method" || meta.kind === "Property" || meta.kind === "Class") {
            const [owner, parents] = getOwners(meta.owner);
            const ownDecorators = getDecorators(owner, meta.kind, meta.name);
            const parentDecorators = parents.map(x => getDecorators(x, meta.kind, meta.kind === "Class" ? x.name : meta.name));
            return Object.assign(Object.assign({}, meta), { decorators: extendDecorators(ownDecorators, flat(parentDecorators)) });
        }
        return meta;
    }
    visitors.addsDecorators = addsDecorators;
    function addsTypeOverridden(meta, ctx) {
        if (meta.kind === "Constructor" || meta.kind === "Class")
            return meta;
        const arrayOverride = meta.decorators.find((x) => x.kind === "Array");
        const override = meta.decorators.find((x) => x.kind === "Override");
        const overridden = (override === null || override === void 0 ? void 0 : override.type) || arrayOverride && [arrayOverride.type];
        if (meta.kind === "Method")
            return Object.assign(Object.assign({}, meta), { returnType: overridden !== null && overridden !== void 0 ? overridden : meta.returnType });
        return Object.assign(Object.assign({}, meta), { type: overridden !== null && overridden !== void 0 ? overridden : meta.type });
    }
    visitors.addsTypeOverridden = addsTypeOverridden;
    function addsGenericOverridden(meta, ctx) {
        /*
        steps
        1. Check if type is of type string
        2. Get the `target` property of the @type decorator. The target property is the parent generic
        3. Get @generic.template of the target property contains list of template type string
        4. Get list of parents classes of `ctx.target`, the order is like: [... GrandParent, Parent, Class]
        5. Get class inherit the generic type (it MUST be the next class of the target - Generic class IS ALWAYS the direct child)
        6. If the inherited class is undefined, its mean the member owned by the generic class itself, return the type (string) immediately
        6. Get @generic.type of the inherited class contains list of types
        7. Create map of <String, Class> from list step 3 and step 6
        8. Resolve the type of step 1 using map from step 7
        */
        if (meta.kind === "Constructor" || meta.kind === "Class")
            return meta;
        const getGenericTypeDecorator = (target) => {
            const decorator = getDecorators(target, "Class", target.name).find((x) => x.kind === "GenericType");
            return decorator === null || decorator === void 0 ? void 0 : decorator.types;
        };
        const getGenericTemplateDecorator = (target) => {
            const decorator = getDecorators(target, "Class", target.name).find((x) => x.kind === "GenericTemplate");
            return decorator === null || decorator === void 0 ? void 0 : decorator.templates;
        };
        const isString = (type) => {
            return typeof type === "string" || (Array.isArray(type) && typeof type[0] === "string");
        };
        const getConversion = (decorators, original) => {
            const decorator = decorators.find((x) => x.kind === "Override");
            if (!decorator)
                return original;
            const rawType = decorator.type;
            if (isString(rawType)) {
                const type = Array.isArray(rawType) ? rawType[0] : rawType;
                const templates = getGenericTemplateDecorator(decorator.target);
                if (!templates)
                    throw new Error(`Configuration Error: ${meta.kind} "${meta.name}" on ${decorator.target.name} uses generic type "${type}" but doesn't specify generic template @generic.template()`);
                const parents = getParents(ctx.target);
                const idx = parents.findIndex(x => x === decorator.target);
                const inherited = parents[idx + 1];
                if (!inherited)
                    return rawType;
                const types = getGenericTypeDecorator(inherited);
                if (!types)
                    throw new Error(`Configuration Error: ${inherited.name} inherited from generic type ${decorator.target.name} but doesn't specify generic type @generic.type()`);
                if (types.length !== templates.length)
                    throw new Error(`Configuration Error: Number of parameter passed on ${decorator.target.name} @generic.template() doesn't match with inherited type ${inherited.name} @generic.type()`);
                const map = new Map(templates.map((x, i) => ([x, types[i]])));
                const conversion = map.get(type);
                return Array.isArray(rawType) ? [conversion] : conversion;
            }
            return rawType;
        };
        if (meta.kind === "Method") {
            const returnType = getConversion(meta.decorators, meta.returnType);
            return Object.assign(Object.assign({}, meta), { returnType });
        }
        const type = getConversion(meta.decorators, meta.type);
        return Object.assign(Object.assign({}, meta), { type });
    }
    visitors.addsGenericOverridden = addsGenericOverridden;
    function addsTypeClassification(meta, ctx) {
        const get = (type) => {
            if (type === undefined)
                return undefined;
            else if (Array.isArray(type))
                return "Array";
            else if (helpers_1.metadata.isCustomClass(type))
                return "Class";
            else
                return "Primitive";
        };
        if (meta.kind === "Method")
            return Object.assign(Object.assign({}, meta), { typeClassification: get(meta.returnType) });
        else if (meta.kind === "Property" || meta.kind === "Parameter")
            return Object.assign(Object.assign({}, meta), { typeClassification: get(meta.type) });
        else if (meta.kind === "Class")
            return Object.assign(Object.assign({}, meta), { typeClassification: "Class" });
        return meta;
    }
    visitors.addsTypeClassification = addsTypeClassification;
    function addsParameterProperties(meta, ctx) {
        if (isParameterProperties(meta) && ctx.parent.kind === "Class") {
            const isParamProp = ctx.parent.decorators.some((x) => x.type === "ParameterProperties");
            return !!isParamProp ? meta : undefined;
        }
        return meta;
    }
    visitors.addsParameterProperties = addsParameterProperties;
    function removeIgnored(meta, ctx) {
        if (meta.kind === "Property" || meta.kind === "Method") {
            const decorator = meta.decorators.find((x) => x.kind === "Ignore");
            return !decorator ? meta : undefined;
        }
        return meta;
    }
    visitors.removeIgnored = removeIgnored;
})(visitors || (visitors = {}));
exports.visitors = visitors;
